SOFTWARE_VERSION_NUMBER = "2.2.0"
# STELLA-2 instrument code
# Science and Technology Education for Land/ Life Assessment
# Paul Mirel 2023

HUGE_VALUE = False #99999
# change HUGE_VALUE to False to allow ordinary measurements of irradiance
MAX_VALUE = 99999 #cap irradiance values at 100,000 - 1 to prevent memory allocation crash

# import system libraries
import gc #garbage collection, RAM management
gc.collect()
print("start memory free {} kB".format( gc.mem_free()/1000 ))
last_alloc = gc.mem_alloc()
import os
import sys
import board
import microcontroller
import time
import rtc
import busio
import digitalio
import terminalio
import storage
import sdcardio
from analogio import AnalogIn

# import device specific libraries
import adafruit_mlx90614    # thermal infrared
import adafruit_mcp9808     # air temperature
import adafruit_as726x      # visible spectrum
import adafruit_pcf8523     # real time hardware_clock
import adafruit_bmp280      # barometer altimeter
import adafruit_sht31d      # humidity
from adafruit_lsm6ds.lsm6ds33 import LSM6DS33   # motion: acceleration and gyroscope
from adafruit_lsm6ds import Rate, AccelRange, GyroRange
import adafruit_lis3mdl     # magnetic field

# import bluetooth libraries
from adafruit_ble import BLERadio # bluetooth
from adafruit_ble.advertising.standard import ProvideServicesAdvertisement
from adafruit_ble.services.nordic import UARTService
from adafruit_bluefruit_connect.packet import Packet
from adafruit_bluefruit_connect.button_packet import ButtonPacket
from adafruit_bluefruit_connect.location_packet import LocationPacket

alloc = gc.mem_alloc()
free = gc.mem_free()
used = alloc - last_alloc
percent_used = 100 * used / ( free + alloc )
print("memory used for library imports, B : {:} which is {:.1f}% of total".format(used, percent_used))

# set constants
UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
print( "UID:", UID )
LOW_BATTERY_VOLTAGE = 3.1
DATA_FILE = "/sd/data.csv"
VIS_BANDS = ( 450, 500, 550, 570, 600, 650 ) # from amd as7262 datasheet
VIS_BAND_PREFIXES = ( "V", "B",  "G", "Y", "O", "R" )
NIR_BANDS = ( 610, 680, 730, 760, 810, 860 ) # from amd as7263 datasheet
MOTIONS = ( "ax", "ay", "az", "gx", "gy", "gz" )
MAG_FIELD = ( "mag_x", "mag_y", "mag_z" )
ON = const(0)  #sample_indicator is active low.
OFF = const(1)

def main():
    # set desired interval between samples
    desired_sample_interval_s = 0.5 # takes ~0.4 to read and record data
    if desired_sample_interval_s < 0.3:
        desired_sample_interval_s = 0.3 # minimum value for STELLA-1
        # Values less than the loop time cause negative wait times,
        # which skips the wait loop where the inputs get interpreted,
        # so the system not longer responds to inputs.

    gc.collect()
    # initialize busses
    spi_bus = initialize_spi()
    i2c_bus = initialize_i2c()
    uart_bus = initialize_uart()
    bluetooth_low_energy = BLERadio()
    bt_uart = UARTService()
    advertisement = ProvideServicesAdvertisement( bt_uart )
    bluetooth_low_energy.start_advertising( advertisement )


    # get hardware real time clock and use it as the source
    # for the microcontroller real time clock
    hardware_clock = initialize_real_time_clock( i2c_bus )
    system_clock = rtc.RTC()
    system_clock.datetime = hardware_clock.datetime

    gc.collect()

    #initializing hardware devices
    # the initialize_x() functions return false if they fail
    # But, do not try to test with "if not some_sensor", use "if some_sensor==False"
    # because you might get False if the object has a len or or bool function (i.e. a list like object)
    battery_monitor = initialize_battery_monitor()
    sd_card = initialize_sd_card( spi_bus )
    sample_indicator = initialize_sample_indicator() #Active low
    sample_indicator_2 = initialize_sample_indicator_2() #Active low
    power_indicator = initialize_power_indicator()   #Active high
    TIR_sensor = initialize_TIR_sensor( i2c_bus )
    AT_sensor = initialize_AT_sensor( i2c_bus )
    BARO_sensor = initialize_BARO_sensor( i2c_bus )
    HUM_sensor = initialize_HUM_sensor( i2c_bus )
    MOT_sensor = initialize_MOT_sensor( i2c_bus )
    MAG_sensor = initialize_MAG_sensor( i2c_bus )
    VIS_sensor = initialize_VIS_sensor( i2c_bus )
    NIR_sensor = initialize_NIR_sensor( uart_bus )
    aux_lamp = initialize_auxilliary_lamp()
    #hardware devices initialization completed

    set_power_indicator( power_indicator, True )
    phone_latitude = 0
    phone_longitude = 0
    phone_altitude = 0

    # initialize inputs
    touch_screen = False
    pushbutton = initialize_pushbutton()

    number_of_record_pause_states = 2 # record_pause_state 0 means the system is recording, 1 means the system is paused
    number_of_source_lamps_states = 2
    number_of_display_states = 1

    record_pause_state = 0
    last_record_pause_state = 0
    screen_record_pause_press_state = True
    screen_record_pause_last_press_state = False
    pushbutton_press_state = False
    pushbutton_last_press_state = False
    press = [0,0,0]

    source_lamps_state = 0
    last_source_lamps_state = 0
    source_lamp_press_state = False
    source_lamps_last_press_state = False

    display_state = 0
    last_display_state = 3 #startup

    batch_start_s = time.monotonic()
    batch_number = update_batch( read_clock( hardware_clock ) )
    startup = True
    inputs = check_inputs( pushbutton, pushbutton_last_press_state )

    loop_count = 0
    bt_report = True #False
    battery_voltage = 0
    BARO_reading = [0,0]
    while True:
        last_sample_time_s = time.monotonic()

        loop_count += 1
        if loop_count > 100000:
            loop_count = 0
        #print("loop_count == {:}".format(loop_count))
        if loop_count % 10 == 0:
            HUM_response = set_HUM_heater( HUM_sensor, True )
            #print(HUM_response)
        else:
            HUM_response = set_HUM_heater( HUM_sensor, False )
            #print(HUM_response)

        if record_pause_state == 0:
            gc.collect()
            #report_memory( "Recording...  Battery voltage == {:}".format( battery_voltage ))#, display_state == {}".format (display_state) )
            if inputs[0] or startup:
                #print( "check inputs 1" )
                inputs = check_inputs( pushbutton, pushbutton_last_press_state )
                # interval == 0.012s
                last_interval_time = time.monotonic()
                #print( "record_pause_state = 0, which means it's recording, so make measurements")
                last_alloc = gc.mem_alloc()/1000.0
                timestamp = read_clock( hardware_clock )
                if last_record_pause_state == 1:
                    batch_start_s = time.monotonic()
                    batch_number = update_batch( timestamp )
                    last_record_pause_state = 0
                iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec)
                #format( iso8601_utc )
                decimal_hour = timestamp_to_decimal_hour( timestamp )
                weekday = timestamp_to_weekday( timestamp )
                interval_time = time.monotonic()
                interval = interval_time - last_interval_time
                #print("interval == {:}".format(interval))
            if inputs[0] or startup:
                #print( "check inputs 2" )
                inputs = check_inputs( pushbutton, pushbutton_last_press_state )
                # interval == 0.004s
                last_interval_time = time.monotonic()
                #read the devices, from main(), since we already initialised the devices by object name
                surface_temperature_C = read_TIR_sensor( TIR_sensor )
                #print("surface temp read completed")
                air_temperature_C = read_AT_sensor( AT_sensor )
                #print("air temp read completed")
                relative_humidity_percent = read_HUM_sensor( HUM_sensor )
                #print("humidity read completed")
                #print("relative_humidity_percent == {:}".format( relative_humidity_percent ))
                BARO_reading = read_BARO_sensor( BARO_sensor )
                #print("barometer read completed")
                barometric_pressure_hPa = BARO_reading[ 0 ]
                altitude_uncalibrated_m = BARO_reading[ 1 ]
                interval_time = time.monotonic()
                interval = interval_time - last_interval_time
                #print("section 2 completed")
                #print("interval == {:}".format(interval))
            if inputs[0] or startup:
                #print( "check inputs 3" )
                inputs = check_inputs( pushbutton, pushbutton_last_press_state )
                #interval == 0.15s
                last_interval_time = time.monotonic()
                visible_irradiance_uW_per_cm_squared = [0,0,0,0,0,0]
                visible_irradiance_uW_per_cm_squared = read_VIS_spectrum( VIS_sensor )
                nir_irradiance_uW_per_cm_squared = [0,0,0,0,0,0]
                nir_irradiance_uW_per_cm_squared = read_NIR_sensor( NIR_sensor )
                motion_XYZ = read_MOT_sensor( MOT_sensor )
                mag_XYZ = read_MAG_sensor( MAG_sensor )
                #print( motion_XYZ )
                main_battery_voltage = check_battery_voltage( battery_monitor )
                #print( "Battery voltage == {:} V".format( main_battery_voltage ))
                #print("\n")
                if main_battery_voltage < LOW_BATTERY_VOLTAGE:
                    low_battery_voltage_notification( power_indicator, bt_uart )
                    print( "                                                     Low battery: V < {:}".format( LOW_BATTERY_VOLTAGE ))
                interval_time = time.monotonic()
                interval = interval_time - last_interval_time
                #print("interval == {:}".format(interval))

            if inputs[0] or startup:
                #print( "check inputs 4" )
                inputs = check_inputs( pushbutton, pushbutton_last_press_state )
                #interval == 0.006s
                last_interval_time = time.monotonic()
                datapoint = [] #reset datapoint string
                datapoint.append( UID )
                datapoint.append( batch_number )
                datapoint.append( "{:},{:},hh.hhhh,{:}".format(weekday, iso8601_utc, decimal_hour ) )
                # name,units,value,calibration_error
                datapoint.append( "surface_temperature,C,{:},1.0".format( surface_temperature_C ) )
                datapoint.append( "air_temperature,C,{:},0.3".format( air_temperature_C ) )
                datapoint.append( "relative_humidity,%,{:},2".format(  relative_humidity_percent ) )
                datapoint.append( HUM_response ) #HUM sensor heater on/ off
                datapoint.append( "air_pressure,hPa,{:},1".format( barometric_pressure_hPa ) )
                datapoint.append( "altitude_uncalibrated,m,{:},1".format( altitude_uncalibrated_m ) )
                # spectrum, wavelength-units, irradiance-units, bandlabel, irradiance, cal-error
                datapoint.append( "visible_spectrum,nm,uW/cm^2,12/100")
                # VIS_BANDS have a color initial prefix
                for i,band in enumerate( VIS_BANDS ):
                    datapoint.append( band )
                    datapoint.append( visible_irradiance_uW_per_cm_squared[ i ] )
                datapoint.append( "near_infrared_spectrum,nm,uW/cm^2,12/100")
                for i,band in enumerate( NIR_BANDS ):
                    datapoint.append( band )
                    datapoint.append( nir_irradiance_uW_per_cm_squared[ i ] )
                datapoint.append( "motion,accel,m/s^2,gyro,rad/s" )
                for i,motion_name in enumerate( MOTIONS ):
                    datapoint.append( motion_name )
                    #print(motion_name)
                    datapoint.append( motion_XYZ[ i ] )
                    #print(motion_XYZ[i])
                datapoint.append( "magnetic_field,uT" )
                for i,mag_name in enumerate( MAG_FIELD ):
                    datapoint.append( mag_name )
                    #print(mag_name)
                    datapoint.append( mag_XYZ[ i ] )
                    #print(mag_XYZ[i])
                datapoint.append( "battery_voltage,V,{:}".format( main_battery_voltage ) )
                datapoint.append( "phone_latitude, {:}, phone_longitude, {:}, phone_altitude, {:}".format( phone_latitude, phone_longitude, phone_altitude ))
                #print( "phone_latitude, {:}, phone_longitude, {:}, phone_altitude, {:}".format( phone_latitude, phone_longitude, phone_altitude ))
                ###
                interval_time = time.monotonic()
                interval = interval_time - last_interval_time
                #print("interval == {:}".format(interval))
            if inputs[0] or startup:
                #print( "check inputs 5" )
                inputs = check_inputs( pushbutton, pushbutton_last_press_state )
                #interval == 0.06s
                last_interval_time = time.monotonic()
                # write to file
                write_line_success = write_data_to_file( UID, timestamp, decimal_hour, batch_number,
                        visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                        surface_temperature_C, air_temperature_C, relative_humidity_percent,
                        barometric_pressure_hPa, altitude_uncalibrated_m,
                        main_battery_voltage, sample_indicator, sample_indicator_2,
                        motion_XYZ[0], motion_XYZ[1], motion_XYZ[2],    #accel
                        motion_XYZ[3], motion_XYZ[4], motion_XYZ[5],    #gyro
                        mag_XYZ[0], mag_XYZ[1], mag_XYZ[2])             #mag

                mirror_data_over_usb(
                    UID, timestamp, decimal_hour, batch_number,
                    visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent
                    )
                gc.collect()
                interval_time = time.monotonic()
                interval = interval_time - last_interval_time
                #print("interval == {:}".format(interval))
                last_record_pause_state = 0

        else:
            #print( "record_pause_state = 1, which means it's paused")
            print("paused...")
            last_record_pause_state = 1


        if bt_report and loop_count % 10/desired_sample_interval_s == 0:
            bt_uart.write( "Press 2 to report, 4 to stop\n" )
            bt_uart.write( "\n" )
            bt_uart.write( "UID: {:} {:} {:04}-{:02}-{:02}  {:02}:{:02}:{:02}Z  Batch:{:}\n".format(UID, weekday, timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min, timestamp.tm_sec, batch_number))
            bt_uart.write( "\n" )
            bt_uart.write( "ATemp = {:.1f}C STemp = {:.1f}C Vbat = {:.1f}V\n".format( air_temperature_C, surface_temperature_C, main_battery_voltage ))
            bt_uart.write( "BP = {:}hPa  Alt = {:}m  RH = {:}%\n".format( barometric_pressure_hPa, altitude_uncalibrated_m, relative_humidity_percent ))
            bt_uart.write( "\n" )
            bt_uart.write( "  ax        ay        az (m/s^2)\n" )
            bt_uart.write( "{:.2f}     {:.2f}     {:.2f}\n".format( motion_XYZ[0], motion_XYZ[1], motion_XYZ[2] ))
            bt_uart.write( "\n" )
            bt_uart.write( "  wx        wy        wz (rad/s)\n" )
            bt_uart.write( "{:.2f}     {:.2f}     {:.2f}\n".format( motion_XYZ[3], motion_XYZ[4], motion_XYZ[5] ))
            bt_uart.write( "\n" )
            bt_uart.write( "  Bx        By        Bz (uT)\n")
            bt_uart.write( "{:.2f}     {:.2f}      {:.2f}\n".format(mag_XYZ[0], mag_XYZ[1], mag_XYZ[2]))
            bt_uart.write( "\n" )
            bt_uart.write( "WL(nm) Ee(uW/cm^2)   WL(nm) Ee(uW/cm^2)\n" )
            for i in range (0, 6):
                bt_uart.write( "{:}    {:}        {:}    {:}\n".format( VIS_BANDS[i], visible_irradiance_uW_per_cm_squared[i], NIR_BANDS[i], nir_irradiance_uW_per_cm_squared[i] ))
                #print(i)
            for i in range (0, 10):
                bt_uart.write( "\n" )
        #bt_up_pressed = False
        #bt_down_pressed = False
        if bt_uart.in_waiting:
            try:
                packet = Packet.from_stream( bt_uart )
            except:
                ValueError
                print("phone GPS ValueError" )
            #print( "packet received" )

            if isinstance( packet, LocationPacket ):
                phone_latitude = packet.latitude
                phone_longitude = packet.longitude
                phone_altitude = packet.altitude

            if isinstance(packet, ButtonPacket):
                if packet.pressed:
                    if packet.button == ButtonPacket.UP:
                        #bt_up_pressed = True
                        record_pause_state = 0
                        bt_report = False
                    elif packet.button == ButtonPacket.DOWN:
                        #bt_down_pressed = True
                        record_pause_state = 1
                        bt_report = False
                    elif packet.button == ButtonPacket.BUTTON_4:
                        bt_report = False
                    elif packet.button == ButtonPacket.BUTTON_2:
                        bt_report = True

        else:
            #print( "skip waiting for packets" )
            #print( "no bt_uart packet in_waiting" )
            pass


        # wait loop
        exit_wait = 0
        remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
        #print( "                                                            remaining_wait_time == {:}".format(remaining_wait_time))
        while remaining_wait_time > 0 and exit_wait == 0:
            inputs = check_inputs( pushbutton, pushbutton_last_press_state )
            no_change = inputs[0]
            #print( "no_change == {:}".format(no_change))

            pushbutton_press_state = inputs[1] #pushbutton_pressed( pushbutton ) #inputs[1]
            #print("pushbutton_press_state == {:}".format(pushbutton_press_state))
            #print("pushbutton_last_press_state == {:}".format(pushbutton_last_press_state))
            if pushbutton_press_state:              # if the button is pushed, whatever the current record pause state is...
                if pushbutton_last_press_state == False: # if the button is just now being pushed and wasn't before; detect rising edge
                    exit_wait = 1                   # don't wait for the interval to finish, go ahead and jump to recording (if it's already in record mode, skip the current interval anyway)
                record_pause_state = 0              # set system to recording
                pushbutton_last_press_state = True  # set the last press state to "pressed"
                #print("pushbutton_last_press_state == {:}".format(pushbutton_last_press_state))
            elif pushbutton_last_press_state:       # if the button goes from last pushed to now not pushed; detect falling edge
                record_pause_state = 1              # set system to pause
                pushbutton_last_press_state = False # set the last press state to "not pressed"
                exit_wait = 1                       # take action to implement pause right now
                # if the button is not pushed and was last not pushed, don't change state
            if record_pause_state == 0 and last_record_pause_state == 1:
                #print( "about to bt_uart write: recording" )
                bt_uart.write("\nRecording: Press DOWN to pause\n(or release mechanical button)\n")
            if record_pause_state == 1 and last_record_pause_state == 0:
                #print( "about to bt_uart write: paused" )
                bt_uart.write("\nPaused: Press UP to record\n(or hold mechanical button)\n")
            last_record_pause_state = record_pause_state

            remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
            #print( "remaining_wait_time == {:}".format(remaining_wait_time))
            time.sleep( 0.05 )
        startup = False
        # return to top of infinite loop

# function definitions below
def mirror_data_over_usb(
    UID, timestamp, decimal_hour, batch_number,
    visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared,
    surface_temperature_C, air_temperature_C, relative_humidity_percent
    ):
    if True: #False:
        gc.collect()
        #print( "memory free immediately before mirror data == {} kB.".format( gc.mem_free()/1000))
        print( "mirror_data", end=", " )
        print( "UID, {}".format( UID ), end=", ")
        print( "batch, {}".format( batch_number ), end=", ")
        print( "year, {}".format( timestamp.tm_year ), end=", ")
        print( "month, {}".format( timestamp.tm_mon ), end=", ")
        print( "day, {}".format( timestamp.tm_mday ), end=", ")
        print( "hour, {}".format( timestamp.tm_hour ), end=", ")
        print( "min, {}".format( timestamp.tm_min ), end=", ")
        print( "sec, {}".format( timestamp.tm_sec ), end=", ")
        print( "dec_hour, {}".format( decimal_hour ), end=", ")
        print( "v450, {}".format( visible_spectrum_uW_per_cm_squared[ 0 ]), end=", ")
        print( "b500, {}".format( visible_spectrum_uW_per_cm_squared[ 1 ]), end=", ")
        print( "g550, {}".format( visible_spectrum_uW_per_cm_squared[ 2 ]), end=", ")
        print( "y570, {}".format( visible_spectrum_uW_per_cm_squared[ 3 ]), end=", ")
        print( "o600, {}".format( visible_spectrum_uW_per_cm_squared[ 4 ]), end=", ")
        print( "r650, {}".format( visible_spectrum_uW_per_cm_squared[ 5 ]), end=", ")
        print( "610, {}".format( nir_spectrum_uW_per_cm_squared[ 0 ]), end=", ")
        print( "680, {}".format( nir_spectrum_uW_per_cm_squared[ 1 ]), end=", ")
        print( "730, {}".format( nir_spectrum_uW_per_cm_squared[ 2 ]), end=", ")
        print( "760, {}".format( nir_spectrum_uW_per_cm_squared[ 3 ]), end=", ")
        print( "810, {}".format( nir_spectrum_uW_per_cm_squared[ 4 ]), end=", ")
        print( "860, {}".format( nir_spectrum_uW_per_cm_squared[ 5 ]), end=", ")
        print( "surface_temp, {}".format( surface_temperature_C ), end=", ")
        print( "air_temp, {}".format( air_temperature_C ), end=", ")
        print( "rel_humidity, {}".format( relative_humidity_percent ))
    return True

def check_inputs( pushbutton, pushbutton_last_press_state ):
    #print( "start check inputs" )
    pushbutton_press_state = pushbutton_pressed( pushbutton )
    if pushbutton_last_press_state == pushbutton_press_state: # and screen_record_pause_last_press_state == press[0] and source_lamps_last_press_state == press[1]:
        no_change = True
    else:
        no_change = False
    #print( "inputs checked" )
    return no_change, pushbutton_press_state

def stall():
    print("intentionally stalled, press return to continue")
    input_string = False
    while input_string == False:
        input_string = input().strip()

def report_memory( message ):
    gc.collect()
    print("UID: {:} Memory B: Used {:} /Free {:} {}".format( UID, gc.mem_alloc(), gc.mem_free(), message))    #last_alloc = gc.mem_alloc()
    #print( "memory used here, bytes: {}".format (gc.mem_alloc() - last_alloc))
    #time.sleep(5)

def initialize_pushbutton():
    pushbutton = digitalio.DigitalInOut( board.D12 )
    pushbutton.direction = digitalio.Direction.INPUT
    pushbutton.pull = digitalio.Pull.UP
    return pushbutton

def pushbutton_pressed( pushbutton ):
    pushbutton_press_state = not pushbutton.value   #active low, so True is notpushed and False is pushed
    return pushbutton_press_state                   #pushbutton_press_state is True if button is being pushed

def initialize_spi():
    try:
        # spi has already been initialized by displayio/adafruit_ili9341, but reinitialize it here anyway
        spi_bus = board.SPI()
        # initialized spi bus
        return spi_bus
    except ValueError as err:
        print( "Error: spi bus fail: {:}".format(err) )
        return False

def initialize_i2c():
    # Initialize the i2c bus at 100kHz, the fastest speed the tir sensor will accommodate.
    # Default i2c speed is 400kHz. At that speed, the TIR sensor will fail to communicate.
    try:
        i2c_bus = busio.I2C( board.SCL, board.SDA, frequency=100000 )
        # initialized i2c bus
        return i2c_bus
    except ValueError as err:
        print( "Error: i2c bus fail: {:}".format(err) )
        return False

def initialize_uart():
    try:
        uart_bus = busio.UART(board.TX, board.RX, baudrate=115200, bits=8, parity=1, stop=1)
        # initialized uart bus
        return uart_bus
    except ValueError as err: # known error behavior
        print( "Error: uart bus fail: {:}".format(err) )
        return False

def initialize_sample_indicator():
    try:
        sample_indicator_LED = digitalio.DigitalInOut( board.A0 )
        sample_indicator_LED.direction = digitalio.Direction.OUTPUT
        sample_indicator_LED.value = True #active low, True is off
        # initialized sample_indicator
        return sample_indicator_LED
    except Exception as err: # TBD: identify some typical errors
        print( "Error: sample indicator initialization failed {:}".format(err) )
        return False

def initialize_sample_indicator_2():
    try:
        sample_indicator_LED = digitalio.DigitalInOut( board.D5 )
        sample_indicator_LED.direction = digitalio.Direction.OUTPUT
        sample_indicator_LED.value = True #active low, True is off
        # initialized sample_indicator
        return sample_indicator_LED
    except Exception as err: # TBD: identify some typical errors
        print( "Error: sample indicator initialization failed {:}".format(err) )
        return False

def initialize_battery_monitor():
    try:
        battery_monitor = AnalogIn(board.VOLTAGE_MONITOR)
        return battery_monitor
    except Exception as err: # TBD: identify some typical errors
        print( "Error: battery monitor initialization fail {:}".format(err) )

def initialize_power_indicator():
    try:
        power_indicator_LED = digitalio.DigitalInOut( board.D13 )
        power_indicator_LED.direction = digitalio.Direction.OUTPUT
        power_indicator_LED.value = True #active high, True is on
        # initialized sample_indicator
        return power_indicator_LED
    except Exception as err: # TBD: identify some typical errors
        print( "Error: power indicator initialization failed {:}".format(err) )
        return False

def set_power_indicator( power_indicator_LED, set_value ):
    try:
        power_indicator_LED.value = set_value
    except Exception as err: # TBD: identify some typical errors
        print( "Error: power indicator set failed {:}".format(err) )

def initialize_real_time_clock( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        rtc = adafruit_pcf8523.PCF8523( i2c_bus )
        # initialized real time hardware_clock
        return rtc
    except ValueError as err: # known error behavior
        print( "Error: real time hardware_clock init fail: {:}".format(err) )
        return False

def initialize_sd_card( spi_bus ):
    if spi_bus == False:
        return False # will fail later if we try use the directory /sd
    # NOTE: pin D10 is in use by the display for SPI D/C, not available for the SD card chip select.
    # Modify the Adalogger Featherwing to use pin D11 for SD card chip select
    sd_cs = board.D11
    try:
        sdcard = sdcardio.SDCard( spi_bus, sd_cs )
        vfs = storage.VfsFat( sdcard )
        storage.mount( vfs, "/sd" )
        # initialized sd card
        return sdcard
    except ValueError as err: # known error behavior
        print( "Error: sd-card init fail: {:}".format(err) )
        return False
    except OSError as err:
        #TBD distinguish between '[Errno 17] File exists' (already mounted), no card and full card conditions.
        # "card not present or card full: [Errno 5] Input/output error" retry?
        print( "Error: sd card fail: card not present or card full: {:} ".format(err) )
        return False
        #TBD catch the "digital pin already in use" error

def update_batch( timestamp ):
    datestamp = "{:04}{:02}{:02}".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday)
    try:
        with open( "/sd/batch.txt", "r" ) as b:
            try:
                previous_batchfile_string = b.readline()
                previous_datestamp = previous_batchfile_string[ 0:8 ]
                previous_batch_number = int( previous_batchfile_string[ 8: ])
            # TBD: catch error when /sd doesn't exist
            except ValueError:
                previous_batch_number = 0
                # corrupted data in batch number file, setting batch to 0
            if datestamp == previous_datestamp:
                # this is the same day, so increment the batch number
                batch_number = previous_batch_number + 1
            else:
                # this is a different day, so start the batch number at 0
                batch_number = 0
    except OSError:
            print( "batch.txt file not found" )
            batch_number = 0
    batch_string = ( "{:03}".format( batch_number ))
    batch_file_string = datestamp + batch_string
    try:
        with open( "/sd/batch.txt", "w" ) as b:
            b.write( batch_file_string )
        # TBD: catch error when /sd doesn't exist
    except OSError as err:
        print("Error: writing batch.txt {:}".format(err) )
        pass
    batch_string = ( "{:}".format( batch_number ))
    return batch_string

def initialize_TIR_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        TIR_sensor = adafruit_mlx90614.MLX90614( i2c_bus )
        # initialized thermal infrared sensor
        return TIR_sensor
    except ValueError as err:
        print( "Error: thermal infrared sensor fail: {:}".format(err) )
        return False

def initialize_AT_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        AT_sensor = adafruit_mcp9808.MCP9808( i2c_bus )
        # initialized air temperature sensor
        return AT_sensor
    except ValueError as err:
        print( "Error: air temperature sensor fail: {:}".format(err) )
        return False

def initialize_auxilliary_lamp():
    try:
        aux_lamp = digitalio.DigitalInOut( board.D2 )
        aux_lamp.direction = digitalio.Direction.OUTPUT
        aux_lamp.value = True #active low, True is off
        # initialized aux lamp
        return aux_lamp
    except Exception as err: # TBD: identify some typical errors
        print( "Error: auxilliary lamp control pin init failed {:}".format(err) )

def initialize_VIS_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        VIS_sensor = adafruit_as726x.AS726x_I2C( i2c_bus )
        print( "\nSet visible spectral sensor to 16x gain and 166 ms integration time, because those are the values used at the factory during the sensor calibration." )
        VIS_sensor.conversion_mode = VIS_sensor.MODE_2
        print( "Default Gain = {}".format( VIS_sensor.gain ))
        VIS_sensor.gain = 16
        print( "Gain Now = {}".format( VIS_sensor.gain ))
        print( "Default Integration Time = {} ms".format( VIS_sensor.integration_time ))
        VIS_sensor.integration_time = 166 # from 1 to 714 = 255 * 2.8ms ms
        print( "Integration Time Now = {} ms".format( VIS_sensor.integration_time ))
        return VIS_sensor
    except ValueError as err:
        print( "Error: visible spectrum sensor fail: {:}".format(err) )
        return False

def initialize_NIR_sensor( uart_bus ):
    if uart_bus == False:
        return False
    uart_bus.write(b"AT\n")
    data = uart_bus.readline() #typical first response from the device is b'ERROR\n', which indicates it is present.
    if data is None:
        print ( "Error: near infrared spectrum sensor fail" )
        return False
    else:
        print( "\nSet near infrared spectral sensor to 16x gain and 166 ms integration time, because those are the values used at the factory during the sensor calibration." )
        # initialized near infrared spectrum sensor
        # check gain setting
        b = b"ATGAIN\n"
        uart_bus.write(b)
        data = str( uart_bus.readline() )[2]
        print( "default gain setting = {}, where 0 is 1x gain, 1 is 3.7x gain, 2 is 16x gain, and 3 is 64x gain".format( data ))

        b = b"ATGAIN=2\n"
        uart_bus.write(b)
        data = str( uart_bus.readline() )[2:4]
        print( "gain setting successful == {}".format( data ))

        b = b"ATGAIN\n"
        uart_bus.write(b)
        data = str( uart_bus.readline() )[2]
        print( "updated gain setting = {}, where 0 is 1x gain, 1 is 3.7x gain, 2 is 16x gain, and 3 is 64x gain".format( data ))

        #check integration time setting
        b = b"ATINTTIME=166\n"
        uart_bus.write(b)
        data = uart_bus.readline()

        b = b"ATINTTIME\n"
        uart_bus.write(b)
        data = str( uart_bus.readline() )[2:6]
        print( "updated integration time setting = {}ms".format( data ))
        # print( "# NIR spectrum default INTTIME (59 * 2.8ms = 165ms): {:}".format(data))
        return uart_bus

def initialize_BARO_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        BARO_sensor = adafruit_bmp280.Adafruit_BMP280_I2C(i2c_bus)
        BARO_sensor.mode = adafruit_bmp280.MODE_NORMAL
        BARO_sensor.standby_period = adafruit_bmp280.STANDBY_TC_250
        return BARO_sensor
    except ValueError as err:
        print( "Error: barometric pressure sensor fail: {:}".format(err) )
        return False

def initialize_HUM_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        HUM_sensor = adafruit_sht31d.SHT31D(i2c_bus)
        return HUM_sensor
    except ValueError as err:
        print( "Error: humidity sensor fail: {:}".format(err) )
        return False

def initialize_MOT_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        MOT_sensor = LSM6DS33( i2c_bus )
        # MOT_sensor.accelerometer_range = AccelRange.RANGE_16G
        print()
        print( "Accelerometer range set to: +/- %d G" % AccelRange.string[ MOT_sensor.accelerometer_range ])
        print( "         Gyro range set to: +/- %d Deg/s" % GyroRange.string[ MOT_sensor.gyro_range ])
        print()
        print( "Accelerometer rate set to: %d HZ" % Rate.string[ MOT_sensor.accelerometer_data_rate ])
        print( "         Gyro rate set to: %d HZ" % Rate.string[ MOT_sensor.gyro_data_rate ])
        print()
        return MOT_sensor
    except ValueError as err:
        print( "Error: motion sensor fail: {:}".format(err) )
        return False

def initialize_MAG_sensor( i2c_bus ):
    if i2c_bus == False:
        return False
    try:
        MAG_sensor = adafruit_lis3mdl.LIS3MDL( i2c_bus )
        return MAG_sensor
    except ValueError as err:
        print( "Error: magnetic field sensor fail: {:}".format(err) )
        return False

def read_clock( rtc ):
    if rtc == False:
        return time.struct_time( (0,1,1,0,0,0,0,1,0) )
    try:
        return rtc.datetime
    except ValueError as err:
        print( "Error: real time clock fail {:}".format(err) )
        return time.struct_time( (0,1,1,0,0,0,0,1,0) )

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return "Error"

def timestamp_to_weekday( timestamp ):
    try:
        clock_days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        weekday = clock_days[ timestamp.tm_wday ]
        return weekday
    except ValueError as err:
        print( "Error: real time hardware_clock fail: {:}".format(err) )
        return "Error"

def read_TIR_sensor( TIR_sensor ):
    if TIR_sensor == False:
        return 0
    decimal_places = 1
    #TBD significant figures
    try:
        surface_temperature_C = TIR_sensor.object_temperature
        surface_temperature_C = round( surface_temperature_C, decimal_places )
        return surface_temperature_C
    except (ValueError, OSError) as err:
        print( "Error: thermal infrared sensor fail: {:}".format(err) )
        return 0

def read_AT_sensor( AT_sensor ):
    if AT_sensor == False:
        return -273
    # mcp9808 datasheet: accuracy +/- 0.25 C
    decimal_places = 1
    #TBD significant figures
    try:
        air_temperature_C = AT_sensor.temperature
        air_temperature_C = round( air_temperature_C, decimal_places )
        return air_temperature_C
    except ValueError as err:
        print( "Error: air temperature sensor fail: {:}".format(err) )
        return -273

def read_HUM_sensor( HUM_sensor ):
    if HUM_sensor == False:
        return -1
    decimal_places = 1
    try:
        relative_humidity_percent = round( HUM_sensor.relative_humidity, decimal_places )
        return relative_humidity_percent
    except ValueError as err:
        print( "Error: humidity sensor fail: {:}".format(err) )
        return -1

def set_HUM_heater( HUM_sensor, value ):
    if HUM_sensor == False:
        return False
    try:
        HUM_sensor.heater = value
        return int( value )
    except ValueError as err:
        print( "Error: humidity sensor fail: {:}".format(err) )
        return False


def read_BARO_sensor( BARO_sensor ):
    if BARO_sensor == False:
        return False
    try:
        #TEMP_reading_C = BARO_sensor.temperature
        #print("temperature == {:}".format(TEMP_reading_C))
        #print()
        #print("before baro read")
        BARO_reading_hpa = round( BARO_sensor.pressure, 1 )
        #print("after barobaro, before alt")
        ALT_reading_m = round( BARO_sensor.altitude, 1 )
        #print("after alt")
        #print()
        return BARO_reading_hpa, ALT_reading_m
    except ValueError as err:
        print( "Error: barometric pressure sensor fail: {:}".format(err) )
        return False

def read_NIR_sensor( NIR_sensor ):
    # returns: R_610nm, S_680nm, T_730nm, U_760nm, V_810nm, W_860nm
    if NIR_sensor == False:
        return [0 for x in NIR_BANDS]
    decimal_places = 1
    b = b"ATCDATA\n"
    NIR_sensor.write(b)
    data = NIR_sensor.readline()
    if data is None:                    #if returned data is of type NoneType
        print( "Alert: Sensor miswired at main instrument board, or not properly modified for uart usage" )
        print( "Enable serial communication via uart by removing solder from the jumpers labeled JP1," )
        print( "and add solder to the jumper labeled JP2 (on the back of the board)" )
        return [0 for x in NIR_BANDS]
    else:
        #print(gc.mem_free())
        datastr = data.decode("utf-8")
        #datastr = "".join([chr(b) for b in data]) # convert bytearray to string
        datastr = datastr.rstrip(" OK\n")
        if datastr == "ERROR":
            print("Error: NIR read")
            return [0 for x in NIR_BANDS]
        # convert to float, and round, in place
        datalist_nir = datastr.split( ", " )
        for n, value in enumerate(datalist_nir): # should be same number as items in NIR_BANDS
            try:
                as_float = float( value )
                value = int( round( as_float, decimal_places ))
                if value > MAX_VALUE:
                    value = MAX_VALUE
            except ValueError:
                print("Failed to convert '{:}' to a float") # only during verbose/development
                value = 0
            # After converting to float, element {:} of the list is: {:0.1f} and is type".format( n, datalist_nir[n] ) )
            datalist_nir[n] = value
    # R_610nm, S_680nm, T_730nm, U_760nm, V_810nm, W_860nm
    if HUGE_VALUE:
        datalist_nir = ( HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE)
    return datalist_nir

def read_VIS_spectrum( spectral_sensor ):
    gc.collect()
    decimal_places = 1
    # returns readings in a tuple, six channels of data
    # Calibration information from AS7262 datasheet:
    # Each channel is tested with GAIN = 16x, Integration Time (INT_T) = 166ms and VDD = VDD1 = VDD2 = 3.3V, TAMB=25°C.
    # The accuracy of the channel counts/μW/cm2 is ±12%.
    # Sensor Mode 2 is a continuous conversion of light into data on all 6 channels
    # 450nm, 500nm, 550nm, 570nm, 600nm and 650nm
    # sensor.violet returns the calibrated floating point value in the violet channel.
    # sensor.raw_violet returns the uncalibrated decimal count value in the violet channel.
    # that syntax is the same for each of the 6 channels

    # NOTE: the library was written for the visible sensor, as7262.
    # The library works exactly the same for the near infrared sensor, as7263,
    # but the band names are called as if the sensor were the visible one.
    try:
        while not spectral_sensor.data_ready:
            time.sleep(0.01)
        initial_read_time_s = time.monotonic()
        violet_calibrated = int( round( spectral_sensor.violet, decimal_places ))
        if violet_calibrated > MAX_VALUE:
            violet_calibrated = MAX_VALUE
        blue_calibrated = int( round( spectral_sensor.blue, decimal_places ))
        if blue_calibrated > MAX_VALUE:
            blue_calibrated = MAX_VALUE
        green_calibrated = int( round( spectral_sensor.green, decimal_places ))
        if green_calibrated > MAX_VALUE:
            green_calibrated = MAX_VALUE
        yellow_calibrated = int( round( spectral_sensor.yellow, decimal_places ))
        if yellow_calibrated > MAX_VALUE:
            yellow_calibrated = MAX_VALUE
        orange_calibrated = int( round( spectral_sensor.orange, decimal_places ))
        if orange_calibrated > MAX_VALUE:
            orange_calibrated = MAX_VALUE
        red_calibrated = int( round( spectral_sensor.red, decimal_places ))
        if red_calibrated > MAX_VALUE:
            red_calibrated = MAX_VALUE
        final_read_time_s = time.monotonic()
        read_time_s = final_read_time_s - initial_read_time_s
        #print( "Sensor read time s = {}".format( read_time_s ))
        spectral_tuple = ( violet_calibrated, blue_calibrated, green_calibrated, yellow_calibrated, orange_calibrated, red_calibrated )
        if HUGE_VALUE:
            spectral_tuple = ( HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE)
        return spectral_tuple
    except ValueError as err:
        print( "Error: spectral sensor fail: {:}".format(err) )
        return ( 0, 0, 0, 0, 0, 0)

def read_MOT_sensor( MOT_sensor ):
    if MOT_sensor == False:
        return False
    try:
        accel_XYZ = MOT_sensor.acceleration
        gyro_XYZ = MOT_sensor.gyro
        motion_XYZ = accel_XYZ + gyro_XYZ
        return motion_XYZ
    except ValueError as err:
        print( "Error: motion sensor fail: {:}".format(err) )
        return False

def read_MAG_sensor( MAG_sensor ):
    if MAG_sensor == False:
        return False
    try:
        mag_XYZ = MAG_sensor.magnetic
        return mag_XYZ
    except ValueError as err:
        print( "Error: magnetic field sensor fail: {:}".format(err) )
        return False

def check_battery_voltage( battery_monitor ):
    if battery_monitor == False:
        return False
    try:
        battery_voltage = round(( battery_monitor.value * 3.3 ) / 65536 * 2, 2 )
        return battery_voltage
    except ValueError as err:
        print( "Error: battery monitor fail: {:}".format(err) )
        return False

def low_battery_voltage_notification( power_indicator, bt_uart ):
    for index in range( 5 ):
        power_indicator.value = False
        time.sleep( 0.05 )
        power_indicator.value = True
        time.sleep( 0.05 )
    bt_uart.write( "Low battery: plug in to recharge\n" )

def write_data_to_file( UID, timestamp, decimal_hour, batch_number,
                        visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                        surface_temperature_C, air_temperature_C, relative_humidity_percent,
                        barometric_pressure_hPa, altitude_uncalibrated_m,
                        main_battery_voltage, sample_indicator, sample_indicator_2,
                        ax, ay, az, gx, gy, gz, mx, my, mz):

    # we let operations fail if the sdcard didn't initialize
    need_header = False
    try:
        os.stat( DATA_FILE ) # fail if data.cav file is not already there
        #raise OSError # uncomment to force header everytime
    except OSError:
        # setup the header for first time
        need_header = True
    try:
        # with sys.stdout as f: # uncomment to write to console (comment "with open")
        with open( DATA_FILE, "a" ) as f: # open seems about 0.016 secs.
            if sample_indicator:
                sample_indicator.value = ON
            if sample_indicator_2:
                sample_indicator_2.value = ON
            if need_header:
                header = (
                    "UID, batch_number, weekday, timestamp, decimal_hour, "
                    "surface_temperature_C, surface_temperature_error_C, "
                    "air_temperature_C, air_temperature_error_C, "
                    "relative_humidity_percent, relative_humidity_error_percent, "
                    "barometric_pressure_hPa, barometric_pressure_error_hPa, "
                    "altitude_uncalibrated_m, altitude_uncalibrated_error_m, "
                    "V450_wavelength_nm, V450_wavelength_error_nm, V450_irradiance_uW_per_cm_squared, V450_irradiance_error_uW_per_cm_squared, "
                    "B500_wavelength_nm, B500_wavelength_error_nm, B500_irradiance_uW_per_cm_squared, B500_irradiance_error_uW_per_cm_squared, "
                    "G550_wavelength_nm, G550_wavelength_error_nm, G550_irradiance_uW_per_cm_squared, G550_irradiance_error_uW_per_cm_squared, "
                    "Y570_wavelength_nm, Y570_wavelength_error_nm, Y570_irradiance_uW_per_cm_squared, Y570_irradiance_error_uW_per_cm_squared, "
                    "O600_wavelength_nm, O600_wavelength_error_nm, O600_irradiance_uW_per_cm_squared, O600_irradiance_error_uW_per_cm_squared, "
                    "R650_wavelength_nm, R650_wavelength_error_nm, R650_irradiance_uW_per_cm_squared, R650_irradiance_error_uW_per_cm_squared, "
                    "nir610_wavelength_nm, nir610_wavelength_error_nm, nir610_irradiance_uW_per_cm_squared, nir610_irradiance_error_uW_per_cm_squared, "
                    "nir680_wavelength_nm, nir680_wavelength_error_nm, nir680_irradiance_uW_per_cm_squared, nir680_irradiance_error_uW_per_cm_squared, "
                    "nir730_wavelength_nm, nir730_wavelength_error_nm, nir730_irradiance_uW_per_cm_squared, nir730_irradiance_error_uW_per_cm_squared, "
                    "nir760_wavelength_nm, nir760_wavelength_error_nm, nir760_irradiance_uW_per_cm_squared, nir760_irradiance_error_uW_per_cm_squared, "
                    "nir810_wavelength_nm, nir810_wavelength_error_nm, nir810_irradiance_uW_per_cm_squared, nir810_irradiance_error_uW_per_cm_squared, "
                    "nir860_wavelength_nm, nir860_wavelength_error_nm, nir860_irradiance_uW_per_cm_squared, nir860_irradiance_error_uW_per_cm_squared, "
                    "battery_voltage, "
                    "ax_m_per_s_squared, ay_m_per_s_squared, az_m_per_s_squared, "
                    "gx_deg_per_s, gy_deg_per_s, gz_deg_per_s, "
                    "mx_uT, my_uT, mz_uT"
                    "\n"
                    )
                f.write( header )
            #for i,v in enumerate(values): # 0.16 secs to write
            #    f.write( str(v) )
            #    if i < len(values)-1: # no trailing ','
            #    f.write(",")
            iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
            weekday = timestamp_to_weekday( timestamp )
            decimal_hour = timestamp_to_decimal_hour( timestamp )
            f.write( "{}, {}, {}, {}, {}, ".format( UID, batch_number, weekday, iso8601_utc, decimal_hour ))
            f.write( "{}, 1.0, ".format( surface_temperature_C ))
            f.write( "{}, 0.3, ".format( air_temperature_C ))
            f.write( "{}, 1.8, ".format( relative_humidity_percent ))
            f.write( "{}, 1, ".format( barometric_pressure_hPa ))
            f.write( "{}, 100, ".format( altitude_uncalibrated_m ))
            for i,band in enumerate( VIS_BANDS ):
                f.write( str(VIS_BANDS[ i ]) )
                f.write( ", ")
                f.write( "5, " )
                f.write( str(visible_irradiance_uW_per_cm_squared[ i ] ))
                f.write( ", ")
                f.write( str(visible_irradiance_uW_per_cm_squared[ i ] * 12/100) )
                f.write( ", ")
            for i,band in enumerate( NIR_BANDS ):
                f.write( str(NIR_BANDS[ i ] ))
                f.write( ", ")
                f.write( "5, " )
                f.write( str(nir_irradiance_uW_per_cm_squared[ i ] ))
                f.write( ", ")
                f.write( str( nir_irradiance_uW_per_cm_squared[ i ] * 12/100 ))
                f.write( ", ")
            f.write( "{}, ".format( main_battery_voltage ) )
            f.write( "{}, ".format( ax ) )
            f.write( "{}, ".format( ay ) )
            f.write( "{}, ".format( az ) )
            f.write( "{}, ".format( gx ) )
            f.write( "{}, ".format( gy ) )
            f.write( "{}, ".format( gz ) )
            f.write( "{}, ".format( mx ) )
            f.write( "{}, ".format( my ) )
            f.write( "{}".format( mz ) )
            f.write("\n")
        if sample_indicator:
            sample_indicator.value = OFF
        if sample_indicator_2:
            sample_indicator_2.value = OFF
        return True
    except OSError as err:
        # TBD: maybe show something on the display like sd_full? this will "Error" every sample pass
        # "[Errno 30] Read-only filesystem" probably means no sd_card
        print( "Error: sd card fail: {:} ".format(err) )
        if sample_indicator or sample_indicator_2:
            sample_indicator.value = ON #  constant ON to show error, likely no SD card present, or SD card full.
            sample_indicator_2.value = ON
        return False

def source_lamps_on(VIS_sensor, NIR_sensor, aux_lamp):
    if aux_lamp != None:
        aux_lamp.value = True # the lamp circuit is active high

    if VIS_sensor != None:
        VIS_sensor.driver_led_current = 12.5 # mA, options 12.5, 25, 50, 100 mA
        VIS_sensor.driver_led = True

    if NIR_sensor != None:
        s = "ATLEDC=48\n"   #0, 16, 32, 48 set the driver to 12.5, 25, 50, 100 mA
        b = bytearray()
        b.extend(s)
        NIR_sensor.write(b)
        #print( "Bytearray sent: %s" % b )
        data = NIR_sensor.readline()
        #print( "Data received: %s" % data)
        s = "ATLED1=100\n" #100 is on, 0 is off
        b = bytearray()
        b.extend(s)
        NIR_sensor.write(b)
        #print( "Bytearray sent: %s" % b )
        data = NIR_sensor.readline()
        #print( "Data received: %s" % data)

def source_lamps_off(VIS_sensor, NIR_sensor, aux_lamp):
    if aux_lamp != None:
        aux_lamp.value = False # the lamp circuit is active high
    if VIS_sensor != None:
        VIS_sensor.driver_led = False
    if NIR_sensor != None:
        s = "ATLED1=0\n" #100 is on, 0 is off
        b = bytearray()
        b.extend(s)
        NIR_sensor.write(b)
        #print( "Bytearray sent: %s" % b )
        data = NIR_sensor.readline()
        #print( "Data received: %s" % data)

main()

